unit oaGauges;

{ (C) 1995, ingenieursbureau Office Automation
  All Rights Reserved

  Hereby the right to distribute this work electronically is granted,
  provided such is done for at most a nominal fee. Also the right is
  granted to store this work on a computer system.
  Finally the right is granted to incorporate this work into other
  work provided no fee is asked for this work.
  In all cases of distribution this work must be distributed in full,
  which specifically includes this notice.
  Liability is limited to the amount payed for this work. Legal
  jurisdiction is with the court of Leeuwarden, the Netherlands.

  Roelof Osinga, 23 october 1995

  PS The price for this work is a toast to my continued good health!

  CompuServe: [100013,262]
  Internet: roelof@iboa.knoware.nl
}
{
  Usage: Like BI's TGauge on which it is, in part, based.

  Bevel: 3D effects for the component
  BevelMargin: between component bevel and gauge
  ProcessMessages: (Win16 only) provide breathing space for Windows
  ProgressBevel: 3D effect for the gauge

  This component is intended to be called within loops, hence the
  ProcessMessages property which will perform an Application.ProcessMessages.
  Optimizations added are the usage of bitmaps that are created once
  for the lifespan of the component. A redraw will only be performed
  if the percentual progress has been changed, i.e. when a change will
  be visual. For large loops this will not occur very often. The other
  changes are icings on the cake like 3D effects possible from within
  the component itself.

  Other important additions are the event handlers. These will allow
  you to respond to mouseclicks and dragging. For example, you could
  popup a menu, stop a process or react to the dropping of a batch
  process's name by starting it up.
}

interface

uses WinTypes, Messages, Classes, Graphics, Controls, Forms, StdCtrls;

type
  TBltBitmap = class(TBitmap)
    procedure MakeLike(ATemplate: TBitmap);
  end;

  TGaugeKind = (gkText, gkHorizontalBar, gkVerticalBar, gkPie, gkNeedle);
  TBevelShape = (bvNone, bvRaised, bvLowered, bvBorder, bvFrame);

  ToaGauge = class(TGraphicControl)
  private
    FMinValue: Longint;
    FMaxValue: Longint;
    FCurValue: Longint;
    FPercent : Integer;
    FKind: TGaugeKind;
    FShowText: Boolean;
    {FBorderStyle: TBorderStyle;}
    FForeColor: TColor;
    FBackColor: TColor;
    FImage : TBitmap;
    FOverlay : TBltBitmap;
    FInitialized : boolean;
  {$IFNDEF WIN32}
    FProcessMessages : boolean; {only for Win16}
  {$ENDIF}
    FBevel : TBevelShape; {broadcast this nonsense}
    FBevelMargin : Integer;
    FProgressBevel : TBevelShape;
    XWidth, XHeight : Integer;
    procedure ImageBackground(AnImage: TBitmap; var aRect : TRect);
    procedure ImageBevel(AnImage: TBitmap; bvKind : TBevelShape; var aRect : TRect);
    procedure OverlayBackground(AnImage: TBitmap; var aRect : TRect);
    procedure PaintAsText(AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsNothing(AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsBar(AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsPie(AnImage: TBitmap; PaintRect: TRect);
    procedure PaintAsNeedle(AnImage: TBitmap; PaintRect: TRect);
    procedure SetBevel(Value : TBevelShape);
    procedure SetBevelMargin(Value : Integer);
    procedure SetProgressBevel(Value : TBevelShape);
    procedure SetGaugeKind(Value: TGaugeKind);
    procedure SetShowText(Value: Boolean);
    {procedure SetBorderStyle(Value: TBorderStyle);}
    procedure SetForeColor(Value: TColor);
    procedure SetBackColor(Value: TColor);
    procedure SetMinValue(Value: Longint);
    procedure SetMaxValue(Value: Longint);
    procedure SetProgress(Value: Longint);
    procedure SetPercent;
    procedure PaintProgress;
    {procedure MouseDown(button : TMouseButton; shift : TShiftState; x, y : integer); override;}
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddProgress(Value: Longint);
  published
    property Align;
    property Color;
    property Enabled;
    property PercentDone: Integer read FPercent;
{$IFNDEF WIN32}
    property ProcessMessage : boolean read FProcessMessages write FProcessMessages default True;
{$ENDIF}
    property Bevel : TBevelShape read FBevel write SetBevel default bvNone;
    property BevelMargin : Integer read FBevelMargin write SetBevelMargin default 0;
    property ProgressBevel : TBevelShape read FProgressBevel write SetProgressBevel default bvNone;
    property Kind: TGaugeKind read FKind write SetGaugeKind default gkHorizontalBar;
    property ShowText: Boolean read FShowText write SetShowText default True;
    property Font;
    {property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;}
    property ForeColor: TColor read FForeColor write SetForeColor default clBlack;
    property BackColor: TColor read FBackColor write SetBackColor default clWhite;
    property MinValue: Longint read FMinValue write SetMinValue default 0;
    property MaxValue: Longint read FMaxValue write SetMaxValue default 100;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property Progress: Longint read FCurValue write SetProgress;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

uses WinProcs, SysUtils;

{ =====================================================
  TBltBitmap
  ===================================================== }

procedure TBltBitmap.MakeLike(ATemplate: TBitmap);
begin
  Width := ATemplate.Width;
  Height := ATemplate.Height;
end;

{ =====================================================
  Solver Functions
  ===================================================== }

{ This function solves for x in the equation "x is y% of z". }
function SolveForX(Y, Z: Longint): Integer;
begin
  SolveForX := Trunc( Z * (Y * 0.01) );
end;

{ This function solves for y in the equation "x is y% of z". }
function SolveForY(X, Z: Longint): Integer;
begin
  if Z = 0
  then SolveForY := 0
  else SolveForY := Trunc( (X * 100) / Z );
end;

{ =====================================================
  ToaGauge
  ===================================================== }

procedure ToaGauge.SetPercent;
var oldPercent : integer;
begin
  oldPercent := FPercent;
  FPercent := SolveForY(FCurValue - FMinValue, FMaxValue - FMinValue);
  if FPercent > 100
  then FPercent := 100
  else
    if FPercent < 0
    then FPercent := 0;
  if FPercent <> oldPercent
  then PaintProgress;
end;

{nonsense to allow run-time toggling of bevels}
{procedure ToaGauge.MouseDown(button : TMouseButton; shift : TShiftState; x, y : integer);
begin
  inherited MouseDown(button, shift, x, y);
  case button of
    mbRight : begin
      if FBevel = high(FBevel)
      then FBevel := bvNone
      else inc(FBevel);
    end;
    mbLeft : begin
      if FProgressBevel = high(FProgressBevel)
      then FProgressBevel := bvNone
      else inc(FProgressBevel);
    end;
  end;
  PaintProgress;
end;}

constructor ToaGauge.Create(AOwner: TComponent);
var bad : boolean;
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csFramed, csOpaque];
  { default values }
  FMinValue := 0;
  FMaxValue := 100;
  FCurValue := 0;
  FPercent := -1;
  FKind := gkHorizontalBar;
  FShowText := True;
  {FBorderStyle := bsSingle;}
  FForeColor := clBlack;
  FBackColor := clWhite;
  FBevel := bvNone;
  FBevelMargin := 0;
  FProgressBevel := bvNone;
  Width := 50;
  Height := 50;
  XWidth := Width;
  XHeight := Height;
  FInitialized := false;
{$IFNDEF WIN32}
  FProcessMessages := true;
{$ENDIF}
  bad := false;
  FImage := TBitmap.Create;
  try
    FImage.Height := Height;
    FImage.Width := Width;
    FOverlay := TBltBitmap.Create;
    try
      FOverlay.MakeLike(FImage);
    except
      on EOutOfMemory do
        bad := true;
      else bad := true;
    end;
    if bad
    then FImage.Free;
  except
    on EOutOfMemory do
    else bad := true;
  end;
  if bad
  then begin
    FImage := nil;
    FOverlay := nil;
  end;
end;

destructor ToaGauge.Destroy;
begin
  FImage.Free;
  FOverlay.Free;
  inherited Destroy;
end;

procedure ToaGauge.Paint;
begin
  if Assigned(FImage)
  then begin
    if Height <> XHeight
    then begin
      FImage.Height := Height;
      FOverlay.Height := Height;
      XHeight := Height;
      FInitialized := false;
    end;
    if Width <> XWidth
    then begin
      FImage.Width := Width;
      FOverlay.Width := Width;
      XWidth := Width;
      FInitialized := false;
    end;
    if not FInitialized
    then begin
      dec(FPercent);
      SetPercent;
      FInitialized := true;
    end;
    Canvas.Draw(0, 0, FImage);
  end;
end;

procedure ToaGauge.OverlayBackground(AnImage: TBitmap; var aRect : TRect);
begin
  with anImage.Canvas do
  begin
    Brush.Style := bsSolid;
    CopyMode := cmBlackness;
    CopyRect(ClipRect, anImage.Canvas, ClipRect);
    CopyMode := cmSrcCopy;
  end
end;

procedure ToaGauge.ImageBevel(AnImage: TBitmap; bvKind : TBevelShape; var aRect : TRect);
begin
  with AnImage.Canvas do
  begin
    Brush.Color := Self.Color;
    FillRect(aRect);
    Brush.Style := bsClear;
    with aRect do
      case bvKind of
        bvBorder : begin
          Pen.Color := clBlack;
          rectangle(left, top, right, bottom);
        end;
        bvRaised : begin
          Pen.Color := clBtnShadow;
          rectangle(left, top, right, bottom);
          Pen.Color := clBtnHighlight;
          moveto(left, bottom - 1);
          lineto(left, top);
          lineto(right - 1, top);
        end;
        bvLowered : begin
          Pen.Color := clBtnHighlight;
          rectangle(left, top, right, bottom);
          Pen.Color := clBtnShadow;
          moveto(left, bottom - 1);
          lineto(left, top);
          lineto(right - 1, top);
        end;
        bvFrame : begin
          brush.style := bsClear;
          Pen.Color := clBtnHighlight;
          rectangle(left + 1, top + 1, right, bottom);
          Pen.Color := clBtnShadow;
          rectangle(left, top, right - 1, bottom - 1);
          InflateRect(aRect, -1, -1);
        end;
      end;
    if bvKind <> bvNone
    then InflateRect(aRect, -1, -1);
  end;
end;

procedure ToaGauge.ImageBackground(AnImage: TBitmap; var aRect : TRect);
begin
  with AnImage.Canvas do
  begin
    Brush.Color := Self.Color;
    FillRect(aRect);
    Brush.Style := bsClear;
    ImageBevel(anImage, FBevel, aRect);
    InflateRect(aRect, -FBevelMargin, -FBevelMargin);
    ImageBevel(anImage, FProgressBevel, aRect);
    Brush.Style := bsSolid;
    Brush.Color := clBlack;
    FillRect(aRect);
  end;
end;

procedure ToaGauge.PaintProgress;
var
  PaintRect: TRect;
begin
  if Assigned(FImage)
  then
    with Canvas do
    begin
      PaintRect := ClientRect;
      ImageBackground(FImage, PaintRect);
      OverlayBackground(FOverlay, PaintRect);
      case FKind of
        gkText: PaintAsNothing(FOverlay, PaintRect);
        gkHorizontalBar, gkVerticalBar: PaintAsBar(FOverlay, PaintRect);
        gkPie: PaintAsPie(FOverlay, PaintRect);
        gkNeedle: PaintAsNeedle(FOverlay, PaintRect);
      end;
      FImage.Canvas.CopyMode := cmSrcInvert;
      FImage.Canvas.Draw(0, 0, FOverlay);
      FImage.Canvas.CopyMode := cmSrcCopy;
      if ShowText
      then PaintAsText(FImage, PaintRect);
      Canvas.CopyMode := cmSrcCopy;
      Canvas.Draw(0, 0, FImage);
    end;
{$IFNDEF WIN32}
  if FProcessMessages
  then Application.ProcessMessages;
{$ENDIF}
end;

procedure ToaGauge.PaintAsText(AnImage: TBitmap; PaintRect: TRect);
var
  S: string;
  X, Y: Integer;
  PaintRectje : TRect;
  temp : TPenMode;
begin
  PaintRectje := ClientRect;
  OverlayBackground(FOverlay, PaintRectje);
  S := Format('%d%%', [PercentDone]);
  with FOverlay.Canvas do
  begin
    Brush.Style := bsClear;
    Font := Self.Font;
    Font.Color := clWhite;
    temp := Pen.Mode;
    Pen.Mode := pmNot;
    with PaintRectje do
    begin
      X := (Right - Left + 1 - TextWidth(S)) div 2;
      Y := (Bottom - Top + 1 - TextHeight(S)) div 2;
    end;
    TextRect(PaintRectje, X, Y, S);
    Brush.Style := bsSolid;
    Pen.Mode := temp;
  end;
  AnImage.Canvas.CopyMode := cmSrcInvert;
  AnImage.Canvas.Draw(0, 0, FOverlay);
  AnImage.Canvas.CopyMode := cmSrcCopy;
end;

procedure ToaGauge.PaintAsNothing(AnImage: TBitmap; PaintRect: TRect);
begin
  with AnImage do
  begin
    Canvas.Brush.Color := BackColor;
    Canvas.FillRect(PaintRect);
  end;
end;

procedure ToaGauge.PaintAsBar(AnImage: TBitmap; PaintRect: TRect);
var
  delta, result : longint;
begin
  with AnImage.Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := BackColor;
    FillRect(PaintRect);
    Pen.Color := ForeColor;
    Pen.Width := 1;
    Brush.Color := ForeColor;
    case FKind of
      gkHorizontalBar:
        with PaintRect do
        begin
          delta := Right - Left;
          delta := SolveForX(PercentDone, delta);
          FillRect(Rect(Left, Top, left + delta, bottom));
        end;
      gkVerticalBar:
        with PaintRect do
        begin
          delta := Bottom - Top;
          result := SolveForX(PercentDone, delta);
          FillRect(Rect(Left, bottom - result, Right, bottom));
        end;
    end;
  end;
end;

procedure ToaGauge.PaintAsPie(AnImage: TBitmap; PaintRect: TRect);
var
  MiddleX, MiddleY: Integer;
  Angle, ca, sa: Double;
  W, H: Integer;
begin
  W := PaintRect.Right - PaintRect.Left;
  H := PaintRect.Bottom - PaintRect.Top;
  if ((w div 2) * 2) = w
  then w := w - 2;
  if ((h div 2) * 2) = w
  then h := h - 2;
  with AnImage.Canvas do
  begin
    Brush.Color := Color;
    FillRect(PaintRect);
    Brush.Color := BackColor;
    Pen.Color := ForeColor;
    Pen.Width := 1;
    with PaintRect do
    begin
      Ellipse(Left, Top, Right, Bottom);
      if PercentDone > 0
      then begin
        Brush.Color := ForeColor;
        MiddleX := left + (W div 2);
        MiddleY := bottom - (H div 2);
        Angle := (Pi * ((PercentDone / 50) + 0.5));
        ca := cos(angle);
        sa := sin(angle);
        Pie(Left, Top, Right, Bottom,
          Round(MiddleX * (1 - Cos(Angle))),
          Round(MiddleY * (1 - Sin(Angle))),
          MiddleX, top );
      end;
    end;
  end;
end;

procedure ToaGauge.PaintAsNeedle(AnImage: TBitmap; PaintRect: TRect);
var
  MiddleX: Integer;
  Angle: Double;
  W, H: Integer;
begin
  with PaintRect do
    with anImage.canvas do
    begin
      W := Right - Left;
      H := Bottom - Top;
      if ((w div 2) * 2) = w
      then w := w - 2;
      if ((h div 2) * 2) = w
      then h := h - 2;
      Brush.Color := Color;
      FillRect(PaintRect);
      Brush.Color := BackColor;
      Pen.Color := ForeColor;
      Pen.Width := 1;
      Pie(left, top, right, top + H * 2, right, Bottom - 1, left, Bottom - 1);
      if PercentDone > 0
      then begin
        Pen.Color := ForeColor;
        MiddleX := W div 2;
        MoveTo(left + MiddleX, Bottom - 1);
        Angle := (Pi * ((PercentDone / 100)));
        LineTo(left + Round(MiddleX * (1 - Cos(Angle))),
          Round(top - 1 + H * (1 - Sin(Angle))));
      end;
    end;
end;

procedure ToaGauge.SetGaugeKind(Value: TGaugeKind);
begin
  if Value <> FKind
  then begin
    FKind := Value;
    PaintProgress; {Refresh;}
  end;
end;

procedure ToaGauge.SetShowText(Value: Boolean);
begin
  if Value <> FShowText
  then begin
    FShowText := Value;
    PaintProgress; {Refresh;}
  end;
end;

{procedure ToaGauge.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle
  then begin
    FBorderStyle := Value;
    Refresh;
  end;
end;}

procedure ToaGauge.SetBevel(Value: TBevelShape);
begin
  if Value <> FBevel
  then begin
    FBevel := Value;
    PaintProgress; {Refresh;}
  end;
end;

procedure ToaGauge.SetBevelMargin(Value: Integer);
begin
  if Value <> FBevelMargin
  then begin
    FBevelMargin := Value;
    PaintProgress; {Refresh;}
  end;
end;

procedure ToaGauge.SetProgressBevel(Value: TBevelShape);
begin
  if Value <> FProgressBevel
  then begin
    FProgressBevel := Value;
    PaintProgress; {Refresh;}
  end;
end;

procedure ToaGauge.SetForeColor(Value: TColor);
begin
  if Value <> FForeColor
  then begin
    FForeColor := Value;
    PaintProgress; {Refresh;}
  end;
end;

procedure ToaGauge.SetBackColor(Value: TColor);
begin
  if Value <> FBackColor
  then begin
    FBackColor := Value;
    PaintProgress; {Refresh;}
  end;
end;

procedure ToaGauge.CMFontChanged(var Message: TMessage);
begin
  inherited;
  PaintProgress;
end;

procedure ToaGauge.CMColorChanged(var Message: TMessage);
begin
  inherited;
  PaintProgress;
end;

procedure ToaGauge.SetMinValue(Value: Longint);
begin
  if Value <> FMinValue
  then begin
    FMinValue := Value;
    SetPercent;
  end;
end;

procedure ToaGauge.SetMaxValue(Value: Longint);
begin
  if Value <> FMaxValue
  then begin
    FMaxValue := Value;
    SetPercent;
  end;
end;

procedure ToaGauge.SetProgress(Value: Longint);
begin
  if (FCurValue <> Value) and (Value >= FMinValue) and (Value <= FMaxValue)
  then begin
    FCurValue := Value;
    SetPercent;
  end;
end;

procedure ToaGauge.AddProgress(Value: Longint);
begin
  Progress := FCurValue + Value;
  SetPercent;
end;

{ =====================================================
  Register
  ===================================================== }

procedure Register;
begin
  RegisterComponents('IBOA', [ToaGauge]);
end;

end.
